""" body.py
 Copyright (C) 1998 Aloril
 
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
"""

import whrandom
from types import *
from copy import copy,deepcopy

from thing import thing
import const,geometry,wtime
from event import *
from mind import mind

class mind_events(thing):
    "process events from mind (check validity etc..)"
    def __init__(self, body):
        thing.__init__(self,body.name)
        self.body=body
        self.mind=body.mind
        self.thing_class={}
    def get_thing_class(self, what):
        try:
            return self.thing_class[what]
        except KeyError:
            try:
                exec("from "+what+" import "+what)
                self.thing_class[what]=eval(what)
            except ImportError:
                self.thing_class[what]=thing
            return self.thing_class[what]
    #event handling: 
    #e is current input event
    #may return new events
    def move_event(self,e):
        e2=copy(e)
        w=e.what
        #source
        source_xyz=w.get_xyz()
        if type(source_xyz)!=TupleType:
            return [event("illegal",what=e,what_desc="source_xyz")]
        source_xyz=tuple(map(float,source_xyz))
        #target
        try:
            target_xyz=e.loc.get_xyz()
        except AttributeError:
            target_xyz=e.loc
            if type(target_xyz)!=TupleType:
                return [event("illegal",what=e,what_desc="target_xyz")]
        target_xyz=tuple(map(float,target_xyz))
        #decide how far
        if geometry.distance(source_xyz,target_xyz)>100.0:
            target=geometry.distance_from_begin(source_xyz,target_xyz,100.0)
        else:
            target=e.loc
        e2.loc=target
        if not e2.target: e2.target=e2.what
        return [e2]
    def cut_event(self,e):
        t=e.what #thing
        return [event("change",target=t,what=t,amount=-0.1)]
    def make_event(self,e):
        what=self.get_thing_class(e.what)
        if e.loc:
            t=apply(what,(e.what,e.what_desc,e.what_desc,0.1,0.0,e.loc))
        else:
            t=apply(what,(e.what,e.what_desc,self.mind,0.1))
        return [event("make", what=t)]
    def change_event(self,e):
        if not e.target: e.target=e.what
        return [e]
    def undefined_event(self,e):
        #just pass others
        return [e]


class body(thing):
    def __init__(self, mind):
        thing.__init__(self,mind.name,mind.__doc__,mind.place,
                       mind.status,mind.age,mind.xyz)
        self.mind=mind
        self.mind.body=self
        self.mevents=mind_events(self)
    #event handling: 
    #e is current input event
    #may return new events
    #for others: see thing.py
    #def destroy_event(self, e):
    #    self.mevents.destroy()
    #    self.mind=None
    def undefined_event(self,e):
        return [e]
    def get_make_event(self, id):
        t=self.find_thing_by_id(id)
        return event("make",what=ething(t.id,name=t.name,type=t.get_type()),
                     loc=t.get_loc(), amount=t.status)
    def find_thing_by_id(self, id):
        t=self.w.id_dict[id]
        if not t:
            #print `self`,id,"is destroyed missing!"
            t=self.w.id_dict['_illegal']
            #import time
            #time.sleep(10)
        return t
    def run_step(self,time,input):
        """run one time step:
           feed input to mind and filter/modify output"""
        self.mind.id=self.id
        output=[]
        #random events
        #death (CHEAT!: should use some chance/time)
        #should take into account age too
        if whrandom.randint(0,const.death_frequency)==0:
            #CHEAT! : leak: add things back to world
            #should be done by world
            e=event("destroy", desc="death", what=self)
            input.append(e)
            output.append(e)
        #birth (CHEAT!: should use some chance/time)
        if whrandom.randint(0,const.birth_frequency)==0:
            #CHEAT!
            if self.age>=const.fertility_age:
                #it gives birth
                #CHEAT!: name should be given by parents!
                #do it when genrally implement talk/rumors
                m=self.mind
                name=self.name+`whrandom.randint(0,9)`
                mnew=mind(name,desc=m.__doc__,place=m.place,
                          xyz=m.xyz,
                          personal_knowledge=deepcopy(m.personal_knowledge),
                          local_knowledge=m.local_knowledge,
                          common_knowledge=m.common_knowledge,
                          personal_goals=m.personal_goals,
                          common_goals=m.common_goals)
                bnew=body(mnew)
                e=event("make", desc="birth", what=bnew)
                input.append(e)
                output.append(e)
        #CHEAT!
        #input: no filtering currently (in future: filter when being in sleep or drunk)
        input2=self.analyse_input(input)
        input2=events_replace_thing_with_id(self,input2)
        if const.use_string_events:
            mind_output=eval(self.mind.run_step(time,`input2`))
        else:
            mind_output=self.mind.run_step(time,input2)
        events_replace_id_with_thing(self,mind_output)
        #process that output to check how legal it is
        output=output+self.mevents.analyse_input(mind_output)
        return output
