""" event.py
 Copyright (C) 1998 Aloril
 
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

------------------------------------------------------------
see event doc string

ething(id,name="foo",type="thing.mind")
event("look",what=ething(id,name="foo",type="thing.mind"))
event("say",what=esay(?,?))

Some examples of messages:
"event('make',what=ething('_farmer',name='farmer',type='thing.mind'),loc=(20, 50, 300))"
"event('destroy', what='_farmer')"
"event('move', what='_farmer', loc=(-50, 50, 200))"
"event('move', what='_axe10', loc='_farmer')"
"event('fire', what='_fire1', loc='_house2', target='_house2')"
"""

import string
from types import *

class event_base:
    def strv(self, var):
        value=self.__dict__[var]
        if value: return self.l.append(" "+var+": "+str(value))
        else: return ""
    def reprv(self, var):
        value=self.__dict__[var]
        if value: self.l.append(var+"="+`value`)
    def __str__(self):
        name=self.__class__.__name__
        var_list=self.var_list
        if len(var_list)==0: return name+"()"
        elif len(var_list)==1: return name+"("+str(getattr(self,var_list[0]))+")"
        self.l=[]
        for var in var_list[1:]:
            self.strv(var)
        return "<"+name+"("+str(getattr(self,var_list[0]))+\
               ","+string.join(self.l,",")+")>\n"
    def __repr__(self):
        name=self.__class__.__name__
        var_list=self.var_list
        if len(var_list)==0: return name+"()"
        elif len(var_list)==1: return name+"("+`getattr(self,var_list[0])`+")"
        self.l=[]
        for var in var_list[1:]:
            self.reprv(var)
        return name+"("+`getattr(self,var_list[0])`+","+string.join(self.l,",")+")"

class event(event_base):
    #__dict__ doesn't define order
    """this is main method of communication between things

       command: type of event: examples: "move", "make", "destroy", "say", "fire"
       source: sender of event
       target: receiver of event
       desc: description of event
       what: to what thing event relates or operates
       what_desc: description of above
       loc: location of event
       loc_desc: description of above
       amount: how big fire, how strong visual signal, how much to change object

       command "make":
         what="thing to make" as string or ething
         optional desc=birth
         optional loc=place for thing to appear (otherwise it appears in what
                      issued event)
         issuers:
           mind: event("make",what="axe", what_desc="ordinary axe")
           body: event("make", desc="birth", what=mnew)
           house: event("make",what=fire)

       command "destroy":
         what=thing to destroy
       
       command "change":
         what=what to change
         amount=amount of change

       command "move":
         what=what to move
         loc=location to move into

       command "cut":
         CHEAT! (will likely change)
         what=what thing is used in cutting

       command "sight":
         what=event (for example move, destroy or make)

       command "say":
         CHEAT! not yet really implemented, what should be an esay
         what=string about what has been said

       command "sound" (not used yet):
         what=event

       command "fire":
         what=fire
         loc=place of fire (for observers)
         target=target of fire
         amount=how big it is

       command "extinguish":
         what=fire to extinguish
         target=same as what

       command "illegal":
         what=event
         what_desc=description about whats wrong: source_xyz, target_xyz

       command "debug":
         what=on or off

       command "NYI" (Not Yet Implemented):
         desc=description about what should have been implemented

       command "imaginary":
         desc=description about what events happened ;-)

       command "empty" (NOP, not really used)

       events by observer.py:

       command "load":
         what=name
         optional loc=debug

       command "step" (run one tick in world)

       command "time" (world returns current time)

    """
    var_list=["command", "source", "target", "desc", "what", "what_desc", 
              "loc", "loc_desc", "amount"]
    is_event=1
    def __init__(self, command, source=None, target=None, desc=None,
                 what=None, what_desc=None,
                 loc=None, loc_desc=None, amount=None):
        self.command=command
        self.source=source
        self.target=target
        self.desc=desc
        self.what=what
        self.what_desc=what_desc
        self.loc=loc
        self.loc_desc=loc_desc
        self.amount=amount
    def copy(self):
        e=event(self.command,
                self.source,
                self.target,
                self.desc,
                self.what,
                self.what_desc,
                self.loc,
                self.loc_desc,
                self.amount)
        try:
            if hasattr(e.what,"is_event"):
                e.what=e.what.copy()
        except AttributeError:
            pass
        return e
    def replace_thing_with_id(self):
        if hasattr(self.source,"is_thing"): self.source=self.source.id
        if hasattr(self.target,"is_thing"): self.target=self.target.id
        if hasattr(self.what,"is_thing"): self.what=self.what.id
        if hasattr(self.loc,"is_thing"): self.loc=self.loc.id
        try:
            if hasattr(self.what,"is_event"):
                self.what=self.what.replace_thing_with_id()
        except AttributeError:
            pass
        return self
    def replace_id_with_thing(self,find_thing):
        if type(self.source)==StringType and self.source[0]=='_':
            self.source=find_thing(self.source)
        if type(self.target)==StringType and self.target[0]=='_':
            self.target=find_thing(self.target)
        if type(self.what)==StringType and self.what[0]=='_':
            self.what=find_thing(self.what)
        if type(self.loc)==StringType and self.loc[0]=='_':
            self.loc=find_thing(self.loc)
        try:
            if hasattr(self.what,"is_event"):
                self.what=self.what.replace_id_with_thing(find_thing)
        except AttributeError:
            pass
        return self

class ething(event_base):
    var_list=["id", "name", "type"]
    def __init__(self, id, name=None, type=None):
        self.id=id
        self.name=name
        self.type=type

def get_event_func(self,e):
    c=e.command+"_event"
    try:
        func=self.event_dict[c]
    except KeyError:
        try:
            func=self.event_dict[c]=getattr(self,c)
        except AttributeError:
            func=self.event_dict[c]=self.undefined_event
    return func

def events_replace_thing_with_id(self, input):
    input2=map(event.copy,input)
    return map(event.replace_thing_with_id,input2)
def events_replace_id_with_thing(self, input):
    return map(event.replace_id_with_thing,
               input,[self.find_thing_by_id]*len(input))
